
#include "../ldr/idaldr.h"
#include <exehdr.h>
#include <queue.hpp>
#include <vector>
#include "dlldr.h"


//--------------------------------------------------------------------------
//
//      check input file format. if recognized, then return 1
//      and fill 'fileformatname'.
//      otherwise return 0
//
int idaapi accept_file(linput_t *li, char fileformatname[MAX_FILE_FORMAT_NAME], int n)
{
	HEADER header;
    if (n)
        return 0;

	qlseek(li, 0, SEEK_SET);

    // Read header.
    if (qlread(li, &header, sizeof(header)) != sizeof(header))
        return 0;

    // Check it's tag.
	if (memcmp(&header.Name, "CCDL", 4) != 0)
        return 0;

    qstrncpy(fileformatname, "ChinaChip DL format", MAX_FILE_FORMAT_NAME);
    set_processor_type("MIPSL", SETPROC_ALL);

    return 1;
}

//--------------------------------------------------------------------------
//
//      load file into the database.
//
void idaapi load_file(linput_t *li, ushort /*neflag*/, const char * /*fileformatname*/)
{
	HEADER                   header;
	std::vector<SYMBOL>      symbols;

	inf.baseaddr = 0;
    inf.lflags |= LFLG_PC_FLAT;

	int nsel = 1;
	uint32 offset_export = 0, offset_import = 0;
	for(int i=0; i<4; i++)
	{
		qlseek(li, sizeof(HEADER)*i, SEEK_SET);

        if (qlread(li, &header, sizeof(HEADER)) != sizeof(HEADER))
            break;

		uint32 adr_start, adr_end, unk;
		segment_t* segm;
		switch(header.Type)
		{
		case 0x8: /* IMPT */
			offset_import = header.Offset;
			break;
		case 0x9: /* EXPT */
			offset_export = header.Offset;
			break;
		case 0x1: /* RAWD */
			adr_start = (*((uint32*)&header.Data[8]));
			adr_end = (*((uint32*)&header.Data[4]));
			unk = (*((uint32*)&header.Data[12]));
			file2base(li, header.Offset, adr_start, adr_end, FILEREG_PATCHABLE);
			set_selector(nsel, 0);
			add_segm(nsel, adr_start, adr_end, "RAWD", CLASS_CODE);
			segm = get_segm_by_sel(nsel);
			if (segm != NULL)
			{
				segm->perm = SEGPERM_EXEC | SEGPERM_WRITE | SEGPERM_READ;
				segm->bitness = 1;
				segm->update();
			}
			nsel++;
			add_long_cmt(adr_start, true,
				     "Section:\n"
                     "  File offset   : %08x\n"
					 "  Start Address : %08x\n"
                     "  End Address   : %08x\n"
                     "  Unknown       : %08x\n",
                     header.Offset, adr_start,
                     adr_end, unk);

			file2base(li, header.Offset+(adr_end-adr_start), adr_end, adr_start+header.Size, FILEREG_PATCHABLE);
			set_selector(nsel, 0);
			add_segm(nsel, adr_end, adr_start+header.Size, "RAWD_2", CLASS_CODE);
			add_long_cmt(adr_end, true,
				     "Section:\n"
                     "  File offset   : %08x\n"
                     "  Start Address : %08x\n"
                     "  End Address   : %08x\n",
                     header.Offset+(adr_end-adr_start), adr_end,
                     adr_end+header.Size);
			nsel++;
			break;
		case 0x10000: /* CCDL */
			add_long_cmt(0, true,
				"Test");
		default:
			break;
		}
    }

	if(offset_import > 0)
	{
		uint32 symbols_total;
		long string_start;

		segment_t* segm;
		qlseek(li, offset_import, SEEK_SET);
		symbols.clear();

		qlread(li, &symbols_total, 4);

		qlseek(li, 12, SEEK_CUR);
		for(uint32 j=0; j<symbols_total; j++)
		{
			SYMBOL cur_symbol;
			qlread(li, &cur_symbol, sizeof(SYMBOL));
			symbols.push_back(cur_symbol);
		}

		string_start = qltell(li);

		for(std::vector<SYMBOL>::iterator k = symbols.begin();
			k != symbols.end(); k++)
		{
			char cur_string[255];
			qlseek(li, string_start+k->String_Offset, SEEK_SET);
			for(int l=0; l<255; l++)
			{
				qlread(li, &cur_string[l], 1);
				if(cur_string[l] == 0)
					break;
			}
			set_selector(nsel, 0);
			add_segm(nsel, k->Address, k->Address+4, "IMPT", "XTRN");
			put_long(k->Address, 0);
			set_name(k->Address, cur_string, SN_PUBLIC|SN_NOWARN);
			nsel++;
		}
	}

	if(offset_export > 0)
	{
		qlseek(li, offset_export);
		symbols.clear();

		uint32 symbols_total;

		qlread(li, &symbols_total, 4);

		qlseek(li, 12, SEEK_CUR);
		for(uint32 j=0; j<symbols_total; j++)
		{
			SYMBOL cur_symbol;
			qlread(li, &cur_symbol, sizeof(SYMBOL));
			symbols.push_back(cur_symbol);
		}

		uint32 string_start = qltell(li);

		for(std::vector<SYMBOL>::iterator k = symbols.begin();
			k != symbols.end(); k++)
		{
			char cur_string[255];
			qlseek(li, string_start+k->String_Offset, SEEK_SET);
			for(int l=0; l<255; l++)
			{
				qlread(li, &cur_string[l], 1);
				if(cur_string[l] == 0)
					break;
			}
			set_selector(nsel, 0);
			add_func(k->Address, 0);
			set_name(k->Address, cur_string, SN_PUBLIC|SN_NOWARN);
			nsel++;
		}
	}
}


//----------------------------------------------------------------------
//
//      LOADER DESCRIPTION BLOCK
//
//----------------------------------------------------------------------
loader_t LDSC =
{
  IDP_INTERFACE_VERSION,
  0,                            // loader flags
//
//      check input file format. if recognized, then return 1
//      and fill 'fileformatname'.
//      otherwise return 0
//
  accept_file,
//
//      load file into the database.
//
  load_file,
//
//	create output file from the database.
//	this function may be absent.
//
  NULL,
//      take care of a moved segment (fix up relocations, for example)
  NULL,
//      initialize user configurable options based on the input file.
  NULL,
};
